# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex import __url__
from nusex.blueprints import with_files
from nusex.blueprints.base import Blueprint

INIT_ATTR_MAPPING = {
    "__productname__": '"PROJECTNAME"',
    "__version__": '"PROJECTVERSION"',
    "__description__": '"PROJECTDESCRIPTION"',
    "__url__": '"PROJECTURL"',
    "__docs__": '"https://PROJECTNAME.readthedocs.io/en/latest"',
    "__author__": '"PROJECTAUTHOR"',
    "__author_email__": '"PROJECTAUTHOREMAIL"',
    "__license__": '"PROJECTLICENSE"',
    "__bugtracker__": '"PROJECTURL/issues"',
    "__ci__": '"PROJECTURL/actions"',
}
PYPROJECT_ATTR_MAPPING = {
    "name": '"PROJECTNAME"',
    "version": '"PROJECTVERSION"',
    "description": '"PROJECTDESCRIPTION"',
    "license": '"PROJECTLICENSE"',
    "authors": '["PROJECTAUTHOR <PROJECTAUTHOREMAIL>"]',
    "maintainers": '["PROJECTAUTHOR <PROJECTAUTHOREMAIL>"]',
    "homepage": '"PROJECTURL"',
    "repository": '"PROJECTURL"',
    "documentation": '"https://PROJECTNAME.readthedocs.io/en/latest"',
}
DOCS_ATTR_MAPPING = {
    "project": '"PROJECTNAME"',
    "copyright": '"PROJECTYEAR, PROJECTAUTHOR"',
    "author": '"PROJECTAUTHOR"',
    "release": "PROJECTNAME.__version__",
}


class PythonBlueprint(Blueprint):
    @with_files("PROJECTNAME/__init__.py")
    def modify_init(self, lines):
        for i, line in enumerate(lines[:]):
            if line.startswith("__"):
                k, v = line.split(" = ")
                v = v.strip('"').strip("'")
                new_v = INIT_ATTR_MAPPING.get(k, v)
                lines[i] = f"{k} = {new_v}"

        return "\n".join(lines)

    @with_files("pyproject.toml")
    def modify_pyproject(self, lines):
        in_tool_poetry = False

        for i, line in enumerate(lines[:]):
            if in_tool_poetry:
                if line.startswith("["):
                    in_tool_poetry = False
                    continue

                try:
                    k, v = line.split(" = ")
                    v = v.strip('"').strip("'")
                    new_v = PYPROJECT_ATTR_MAPPING.get(k, v)
                    lines[i] = f"{k} = {new_v}"
                except ValueError:
                    ...

            elif line.strip() == "[tool.poetry]":
                in_tool_poetry = True

        return "\n".join(lines).replace(self.project_name, "PROJECTNAME")

    @with_files("docs/conf.py", "docs/source/conf.py")
    def modify_docs_conf(self, lines):
        in_project_info = False

        for i, line in enumerate(lines[:]):
            if in_project_info:
                if line.startswith("# --"):
                    in_project_info = False
                    continue

                try:
                    k, v = line.split(" = ")
                    v = v.strip('"').strip("'")
                    new_v = DOCS_ATTR_MAPPING.get(k, v)
                    lines[i] = f"{k} = {new_v}"
                except ValueError:
                    ...

            elif line.startswith("# -- Project information"):
                in_project_info = True

            elif line.strip() == f"import {self.project_name}":
                lines[i] = "import PROJECTNAME"

        return "\n".join(lines)

    @with_files("PROJECTNAME/error.py", "PROJECTNAME/errors.py")
    def modify_error_files(self, lines):
        for line in lines[:]:
            if line.startswith("class"):
                base_exc = line.split("(")[0][6:]
                break

        return "\n".join(lines).replace(base_exc, "PROJECTBASEEXC")

    @with_files("README.md", "README.txt")
    def modify_readme(self, lines):
        last_line = len(lines) - 1
        found_acks = False
        ack = (
            "This project was created in part by the [nusex project "
            f"templating utility]({__url__})."
        )

        for i, line in enumerate(lines[:]):
            if line.startswith("#"):
                if found_acks:
                    lines.insert(i, ack)
                    lines.insert(i + 1, "")
                    break

                if "acknowledgements" in line.lower():
                    found_acks = True

            elif i == last_line and found_acks:
                lines.extend([ack, ""])

        if not found_acks:
            lines.extend(["## Acknowledgements", "", ack, ""])

        return "\n".join(lines).replace(self.project_name, "PROJECTNAME")

    @with_files("LICENSE", "LICENSE.txt", "COPYING", "COPYING.txt")
    def modify_license(self, _):
        return "LICENSEBODY"

    @with_files("MANIFEST.in", "setup.cfg", "setup.py")
    def modify_other_files(self, lines):
        # TODO: Make the setup files more complete:
        # https://docs.python.org/3/distutils/setupscript.html
        return "\n".join(lines).replace(self.project_name, "PROJECTNAME")



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from nusex import TEMPLATE_DIR, Template
from nusex.errors import AlreadyExists
from nusex.helpers import cprint


def _options_as_set(values):
    s = set(values.split(","))

    if s == {""}:
        return {}

    return s


def _options_as_list(values):
    l = values.split(",")

    if l == [""]:
        return []

    return l


def _check(template):
    cprint("inf", "Showing template manifest (incl. changes):")
    manifest = template.check()
    for file, meta in manifest.items():
        print(file)
        max_meta = len(meta) - 1
        for i, (ln, line) in enumerate(meta):
            if i == max_meta:
                print(f"└── Line {ln}: {line}")
            else:
                print(f"├── Line {ln}: {line}")

    installs = template.data["installs"]
    if installs:
        print()
        cprint("inf", "Showing dependencies:")
        max_dep = len(installs) - 1
        for i, dep in enumerate(installs):
            if i == max_dep:
                print(f"└── {dep}")
            else:
                print(f"├── {dep}")
        return


def run(
    name,
    overwrite,
    check,
    from_repo,
    with_installs,
    with_requirements_file,
    ignore_exts,
    extend_ignore_exts,
    ignore_dirs,
    extend_ignore_dirs,
):
    if os.path.isfile(TEMPLATE_DIR / f"{name}.nsx") and not overwrite:
        raise AlreadyExists(
            "That template already exists (use -o to overwrite)"
        )

    ignore_exts = ignore_exts.union(extend_ignore_exts)
    ignore_dirs = ignore_dirs.union(extend_ignore_dirs)

    if with_requirements_file:
        with open(with_requirements_file) as f:
            d = f.read().split("\n")
            d.remove("")
            with_installs.extend(d)

    # TODO: Make so when overwriting a template, it doesn't have to
    # load the previous one first.
    if from_repo:
        template = Template.from_repo(
            name,
            from_repo,
            installs=with_installs,
            ignore_exts=ignore_exts,
            ignore_dirs=ignore_dirs,
        )
    else:
        template = Template.from_cwd(
            name,
            installs=with_installs,
            ignore_exts=ignore_exts,
            ignore_dirs=ignore_dirs,
        )

    if check:
        return _check(template)

    template.save()
    cprint("aok", f"Template '{name}' built successfully!")


def setup(subparsers):
    s = subparsers.add_parser(
        "build",
        description="Build a new template.",
    )
    s.add_argument("name", help="the name for the new template")
    s.add_argument(
        "-o",
        "--overwrite",
        help="overwrite an existing template should it already exist",
        action="store_true",
    )
    s.add_argument(
        "-c",
        "--check",
        help="check the build manifest without building the template",
        action="store_true",
    )
    s.add_argument(
        "-r",
        "--from-repo",
        help=(
            "a repository URL to build a template from (Git must be "
            "installed)"
        ),
        metavar="URL",
        default="",
    )
    s.add_argument(
        "-i",
        "--with-installs",
        help=(
            "a comma-separated list of dependencies to install when deploying"
        ),
        metavar="DEPS",
        default="",
        type=_options_as_list,
    )
    s.add_argument(
        "-I",
        "--with-requirements-file",
        help=(
            "a file within the template to install dependencies from when "
            "deploying"
        ),
        metavar="FILENAME",
        default="",
    )
    s.add_argument(
        "--ignore-exts",
        help=(
            "a comma-separated list of file extensions to ignore (default: "
            "pyc,pyd,pyo)"
        ),
        metavar="EXTS",
        default="pyc,pyd,pyo",
        type=_options_as_set,
    )
    s.add_argument(
        "--extend-ignore-exts",
        help=(
            "a comma-separated list of file extensions to ignore on top of "
            "the defaults"
        ),
        metavar="EXTS",
        default="",
        type=_options_as_set,
    )
    s.add_argument(
        "--ignore-dirs",
        help=(
            "a comma-separated list of directories to ignore; look at the "
            "docs for more information about advanced ignoring syntaxes "
            "(default: .direnv,.eggs,.git,.hg,.mypy_cache,.nox,.tox,.venv,"
            "venv,.svn,_build,build,dist,buck-out,.pytest-cache,*.egg-info)"
        ),
        metavar="DIRS",
        default=(
            ".direnv,.eggs,.git,.hg,.mypy_cache,.nox,.tox,.venv,venv,.svn,"
            "_build,build,dist,buck-out,.pytest-cache,*.egg-info"
        ),
        type=_options_as_set,
    )
    s.add_argument(
        "--extend-ignore-dirs",
        help=(
            "a comma-separated list of directories to ignore on top of the "
            "defaults"
        ),
        metavar="DIRS",
        default="",
        type=_options_as_set,
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class NusexError(Exception):
    """The base exception class for nusex."""


class NusexUserError(NusexError):
    """An error occurred due to user error."""


class ProfileError(NusexUserError):
    """Something went wrong when working with a profile."""


class TemplateError(NusexUserError):
    """Something went wrong when working with a template."""


class BuildError(TemplateError):
    """A template failed to build."""


class DeploymentError(TemplateError):
    """A template failed to deploy."""


class AlreadyExists(NusexUserError):
    """A profile or template already exists."""


class DoesNotExist(NusexUserError):
    """A profile or template does not exist."""


class MigrationError(NusexUserError):
    """A migration or migration reversion failed."""


class DownloadError(NusexError):
    """A download failed to complete, or an invalid download request
    was submitted."""


class UnsupportedFile(NusexError):
    """An invalid file has been passed to a decoder, or a file larger
    than the file size limit has been included in a template."""


class IncompatibilityError(NusexError):
    """A operation is incompatible with your OS, Python version, or
    Python implementation."""



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import sys
from pathlib import Path


def _suffix():
    # Determine whether this is a production copy or not. This
    # prevents actual user configs from getting messed up.
    parts = Path(__file__).parts
    if "nox" in " ".join(sys.argv) or ".nox" in parts:
        return "nusex-test"
    if "site-packages" not in parts:
        return "nusex-dev"
    return "nusex"


if os.name == "nt":
    CONFIG_DIR = Path.home() / f".{_suffix()}"
    TEMP_DIR = CONFIG_DIR / "tmp"
else:
    CONFIG_DIR = Path.home() / f".config/{_suffix()}"
    TEMP_DIR = Path(f"/tmp/{_suffix()}")

CONFIG_FILE = CONFIG_DIR / "config.nsc"
LICENSE_DIR = CONFIG_DIR / "licenses"
PROFILE_DIR = CONFIG_DIR / "profiles"
TEMPLATE_DIR = CONFIG_DIR / "templates"

INVALID_NAME_PATTERN = re.compile("[^a-z0-9_]")
# https://github.com/pypa/packaging/blob/16.7/packaging/version.py#L159
VERSION_PATTERN = re.compile(
    r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
""",
    re.VERBOSE | re.IGNORECASE,
)



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
from urllib import request
from urllib.error import HTTPError

from nusex import CONFIG_DIR
from nusex.errors import DownloadError
from nusex.helpers import cprint

REPO_URL = "https://github.com/nusex/downloads"
RAW_URL = "https://raw.githubusercontent.com"
TEMPLATE_SUFFIX = "tree/main/templates1x"
LICENSE_URL = (
    "https://github.com/github/choosealicense.com/tree/gh-pages/_licenses"
)


class Downloader:
    """A utility class for downloading assets.

    Args:
        of_type(str): The type of asset to download (must be either
            "templates" or "licenses").

    Attributes:
        of_type(str): The type of asset to download (must be either
            "templates" or "licenses").
        extension (str): The file extension of the files to search for.
        url (str): The repo to search for files in.
        files (list[str]): A series of file URLs.
        completed (int): The number of files that have been downloaded.
    """

    __slots__ = ("of_type", "extension", "url", "files", "completed")

    def __init__(self, of_type):
        if of_type not in ("templates", "licenses"):
            return DownloadError("You can only download templates or licenses")

        self.of_type = of_type
        if of_type == "templates":
            self.extension = "nsx"
            self.url = f"{REPO_URL}/{TEMPLATE_SUFFIX}"
        else:
            self.extension = "txt"
            self.url = LICENSE_URL
        self.files = []
        self.completed = 0

    @property
    def progress(self):
        """The download progress as a percentage.

        Returns:
            float
        """
        return (100 / len(self.files)) * self.completed

    def fetch(self):
        """Fetch the files to download.

        Raises:
            DownloadError: There was a problem fetching the files.
        """
        self.files = []

        try:
            with request.urlopen(self.url) as r:
                data = r.readlines()
        except HTTPError as exc:
            raise DownloadError(
                f"Fetch failed (GitHub returned {exc.code})"
            ) from None

        for i, line in enumerate(data):
            if b'role="rowheader"' in line:
                path = data[i + 1].split(b'"')[-2].decode()
                if path.endswith(f".{self.extension.lstrip('.')}"):
                    self.files.append("".join(path.split("blob/")))

    async def _download_files(self):
        try:
            for f in self.files:
                with request.urlopen(f"{RAW_URL}/{f}") as r:
                    with open(
                        (CONFIG_DIR / self.of_type) / f.split("/")[-1], "wb"
                    ) as f:
                        f.write(r.read())
                        self.completed += 1

                await asyncio.sleep(0)

        except HTTPError as exc:
            raise DownloadError(
                f"Download failed (GitHub returned {exc.code})"
            ) from None

    async def _display_progress(self):
        while True:
            cprint(
                "prc",
                f"Downloading {self.of_type}... {self.progress:,.0f}%",
                end="\r",
            )
            if self.progress == 100:
                return
            await asyncio.sleep(0)

    def download(self, display_progress=False):
        """Download files. The files will be fetched automatically if
        they have not already been.

        Args:
            display_progress (bool): Whether to display the download's
                progress in the terminal. Defaults to False.

        Raises:
            DownloadError: There was a problem fetching the files.
        """
        if not self.files:
            self.fetch()

        loop = asyncio.get_event_loop()
        download = loop.create_task(self._download_files())
        if display_progress:
            display = loop.create_task(self._display_progress())
            display.add_done_callback(lambda x: print())
        loop.run_until_complete(download)



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex.helpers import cprint
from nusex.spec import NSCSpecIO


def run(auto_update):
    data = NSCSpecIO().read()
    data["auto_update"] = auto_update
    NSCSpecIO().write(data)

    cprint("aok", "Configuration updated!")


def setup(subparsers):
    s = subparsers.add_parser(
        "config", description="Update your nusex configuration."
    )
    s.add_argument(
        "-u",
        "--auto-update",
        help=(
            "activate auto-updating (omitting this flag will deactivate "
            "auto-updating)"
        ),
        action="store_true",
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os

from nusex import CONFIG_DIR, LICENSE_DIR, PROFILE_DIR, VERSION_PATTERN
from nusex.errors import ProfileError
from nusex.helpers import cprint, validate_name
from nusex.spec import NSCSpecIO, NSPSpecIO

VALID_CONFIG_KEYS = (
    "author_name",
    "author_email",
    "git_profile_url",
    "starting_version",
    "default_description",
    "preferred_license",
)


class Profile:
    """A class in which to create, load, modify, and save profiles.

    Keyword Args:
        name (str): The name of the profile. If the profile does not
            exist, a new one is created, otherwise an existing one is
            loaded. Defaults to "default".

    Attributes:
        path (pathlib.Path): The complete filepath to the profile.
        data (dict[str, Any]): The data for the profile.
    """

    __slots__ = ("path", "data")

    def __init__(self, name="default"):
        self.path = PROFILE_DIR / f"{name}.nsp"

        if not os.path.isfile(self.path):
            return self.create_new(name)

        self.load()

    def __str__(self):
        return self.path.stem

    def __repr__(self):
        return f"<Profile name={self.name!r}>"

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __getitem__(self, key):
        return self.data[key]

    @property
    def name(self):
        """The name of this profile.

        Returns:
            str
        """
        return self.path.stem

    @property
    def exists(self):
        """Whether this profile exists on disk.

        Returns:
            bool
        """
        return self.path.is_file()

    def create_new(self, name):
        """Create a new profile.

        Args:
            name (str): The name of the profile.

        Raises:
            ProfileError: The provided name is invalid.
            AlreadyExists: The profile already exists on disk.
        """
        validate_name(name, self.__class__.__name__)
        self.data = {
            "author_name": "",
            "author_email": "",
            "git_profile_url": "",
            "starting_version": "0.1.0",
            "default_description": "My project, created using nusex",
            "preferred_license": "unlicense",
        }

    def load(self):
        """Load an existing profile. This should never need to be called
        as profiles are loaded automatically when necessary upon object
        creation.

        Raises:
            FileNotFoundError: The profile does not exist on disk.
        """
        self.data = NSPSpecIO().read(self.path)

    def save(self):
        """Save this profile.

        Raises:
            ProfileError: The profile data has been improperly modified.
        """
        try:
            NSPSpecIO().write(self.path, self.data)
        except KeyError:
            raise ProfileError(
                "The profile data has been improperly modified"
            ) from None

    def delete(self):
        """Delete this profile.

        Raises:
            FileNotFoundError: The profile does not exist on disk.
        """
        os.remove(self.path)

    def rename(self, new_name):
        """Rename this profile.

        Args:
            new_name (str): The new name for the profile.

        Raises:
            FileNotFoundError: The profile does not exist on disk.
        """
        new_path = f"{self.path}".replace(self.path.stem, new_name)
        self.path = self.path.rename(new_path)

    @classmethod
    def current(cls):
        """Create an instance for the currently selected profile.

        Returns:
            Profile: The currently selected profile.
        """
        return cls(NSCSpecIO().read()["profile"])

    @classmethod
    def from_legacy(cls, name="default"):
        """Create a profile from a 0.x spec user.nsc file.

        Keyword Args:
            name (str): The name of the profile. Defaults to "default".

        Returns:
            Profile: The newly created profile.

        Raises:
            FileNotFoundError: No user.nsc file exists in the config
                directory.
        """
        with open(CONFIG_DIR / "user.nsc") as f:
            data = json.load(f)

        c = cls(name)
        values = list(data.values())
        order = (3, 4, 2, 0, 1, 5)

        for i, k in enumerate(c.data.keys()):
            v = values[order[i]]

            if k == "starting_version":
                if v != "CALVER" and not VERSION_PATTERN.match(v):
                    v = (
                        input(f"🎤 Starting version [0.1.0]: ").strip()
                        or "0.1.0"
                    )
                    v = c._validate_option(k, v)

            elif k == "preferred_license":
                v = c._resolve_license(v)
                if not v:
                    v = (
                        input(f"🎤 Preferred license [unlicense]: ").strip()
                        or "unlicense"
                    )
                    v = c._validate_option(k, v)

            c.data[k] = v

        return c

    @property
    def is_selected(self):
        """Whether this profile is currently selected.

        Returns:
            bool
        """
        return NSCSpecIO().read()["profile"] == self.path.stem

    def select(self):
        """Select this profile."""
        data = NSCSpecIO().read()
        data["profile"] = self.path.stem
        NSCSpecIO().write(data)

    def _resolve_license(self, value):
        for file in LICENSE_DIR.glob("*.txt"):
            if value == file.stem:
                return value

            attrs = {}
            with open(LICENSE_DIR / file, encoding="utf-8") as f:
                # This needs to be explicitly encoded as UTF-8 otherwise
                # it won't work on Windows.
                for line in f:
                    if line == "---\n":
                        continue
                    if line == "\n":
                        break

                    k, v = line.split(": ")
                    attrs.update({k: v.strip()})

            # Provide as many chances for a match as possible.
            if (
                attrs.get("nickname", "") == value
                or attrs.get("spdx-id", "") == value
                or attrs.get("title", "") == value
                or (
                    attrs.get("title", "")
                    .replace("License", "")
                    .replace("  ", " ")
                    .strip()
                )
                == value
            ):
                return file.stem

    def _validate_option(self, key, option):
        if key == "git_profile_url":
            option = option.strip("/")

        if key == "starting_version":
            if option != "CALVER" and not VERSION_PATTERN.match(option):
                raise ProfileError(
                    "That version number does not conform to PEP 440 "
                    "standards, or is not 'CALVER'"
                )

        if key == "preferred_license":
            option = self._resolve_license(option)
            if not option:
                raise ProfileError(
                    "Your input could not be resolved to a valid license"
                )

        return option

    def setup(self):
        """Set up this profile. You will be prompted to input some
        information.

        Raises:
            ProfileError: An invalid value was provided to one of the
                inputs.
        """
        for k, v in self.data.items():
            kq = (k[0].upper() + k[1:].replace("_", " ")).replace("url", "URL")
            option = input(f"🎤 {kq} [{v}]: ").strip() or v.strip()
            option = self._validate_option(k, option)
            self.data[k] = option

    def update(self, **kwargs):
        """Update this profile's configuration. All options must be
        passed through as kwargs. Invalid configuration keys raise
        warnings, not errors.

        Keyword Args:
            author_name (str): An author name.
            author_email (str): An author email.
            git_profile_url (str): Your GitHub/Gitlab/BitBucket/etc.
                profile link.
            starting_version (str): The version to initialise project
                with.
            default_description (str): The description to initialise
                projects with.
            preferred_license (str): Your preferred license.
        """
        for k, v in kwargs.items():
            if k not in VALID_CONFIG_KEYS:
                cprint("war", f"'{k}' is not a valid key, skipping...")
                continue

            if v:
                v = self._validate_option(k, v)
                self.data[k] = v



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import subprocess as sp
import sys

from . import INVALID_NAME_PATTERN, PROFILE_DIR, TEMPLATE_DIR
from .errors import AlreadyExists, ProfileError, TemplateError

MESSAGE_TYPES = {
    "aok": ("🎉", "\33[92m"),
    "war": ("💣", "\33[93m"),
    "err": ("💥", "\33[91m"),
    "prc": ("⌛", ""),
    "inf": ("📢", "\33[94m"),
}


def cprint(type, text, **kwargs):
    emoji, colour = MESSAGE_TYPES[type]
    print(f"{emoji} {colour}{text}\33[0m", **kwargs)


def validate_name(name, for_type):
    err = {
        "Profile": ProfileError,
        "Template": TemplateError,
    }[for_type]

    if len(name) > 24:
        raise err(f"{for_type} names are limited to 24 characters")

    if INVALID_NAME_PATTERN.search(name):
        raise err(
            f"{for_type} names can only contain lower case letters, numbers, "
            "and underscores"
        )

    if name.startswith("nsx"):
        raise err("That name is reserved")

    in_dir = {
        "Profile": TEMPLATE_DIR,
        "Template": PROFILE_DIR,
    }[for_type]

    if name in (f.split(".")[0] for f in os.listdir(in_dir)):
        raise AlreadyExists(
            f"A {'profile' if for_type == 'Template' else 'template'} is "
            "already using that name"
        )


def run(command):
    if sys.version_info >= (3, 7, 0):
        return sp.run(command, shell=True, capture_output=True)

    if os.name != "nt":
        return sp.run(f"{command} > /dev/null 2>&1", shell=True)

    # Windows users will have to put up with the output for 3.6 tests.
    return sp.run(command, shell=True)



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from nusex import PROFILE_DIR, Profile
from nusex.errors import AlreadyExists, DoesNotExist
from nusex.helpers import cprint


def _build_profile(name):
    if os.path.isfile(PROFILE_DIR / f"{name}.nsp"):
        raise AlreadyExists("A profile with that name already exists")

    profile = Profile(name)
    profile.setup()
    profile.save()
    profile.select()
    cprint("aok", f"Profile '{profile.name}' successfully created!")


def run(show_current, create_new, switch, **kwargs):
    if show_current:
        return print(Profile.current().name)

    if create_new:
        return _build_profile(create_new)

    if switch:
        profile = Profile(switch)
        if not profile.exists:
            raise DoesNotExist(f"Profile '{switch}' not found")
        profile.select()
        return cprint("aok", f"Switched to profile '{profile}'!")

    profile = Profile.current()

    if any(kwargs.values()):
        profile.update(**kwargs)
        profile.save()
        return cprint("aok", f"Profile '{profile}' updated!")

    profile.setup()
    profile.save()
    cprint("aok", f"Profile '{profile}' updated!")


def setup(subparsers):
    s = subparsers.add_parser(
        "profile",
        description="Create, modify and switch user profiles.",
    )
    s.add_argument(
        "-c",
        "--show-current",
        help="show the currently selected profile and exit",
        action="store_true",
    )
    s.add_argument(
        "-n",
        "--create-new",
        help="create a new profile",
        metavar="NAME",
        default="",
    )
    s.add_argument(
        "-s",
        "--switch",
        help="switch to a different profile",
        metavar="NAME",
        default="",
    )
    s.add_argument(
        "-a",
        "--author-name",
        help="change the author name for this profile",
        metavar="NAME",
        default="",
    )
    s.add_argument(
        "-e",
        "--author-email",
        help="change the author email for this profile",
        metavar="EMAIL",
        default="",
    )
    s.add_argument(
        "-g",
        "--git-profile-url",
        help="change the Git profile URL for this profile",
        metavar="URL",
        default="",
    )
    s.add_argument(
        "-v",
        "--starting-version",
        help="change the starting version for this profile",
        metavar="VERSION",
        default="",
    )
    s.add_argument(
        "-d",
        "--default-description",
        help="change the default description for this profile",
        metavar="DESCRIPTION",
        default="",
    )
    s.add_argument(
        "-l",
        "--preferred-license",
        help="change the preferred license for this profile",
        metavar="LICENSE",
        default="",
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime as dt
import os
import sys

from nusex import CONFIG_DIR, CONFIG_FILE, Profile
from nusex.helpers import cprint
from nusex.spec import NSCSpecIO
from nusex.utils import Downloader

DIRS = ("licenses", "profiles", "templates")


def run():
    if os.path.isfile(CONFIG_FILE):
        cprint("err", "You've already initialised nusex!")
        sys.exit(2)

    for d in DIRS:
        os.makedirs(CONFIG_DIR / d, exist_ok=True)

    for t in ("templates", "licenses"):
        if not os.listdir(CONFIG_DIR / t):
            Downloader(t).download(display_progress=True)

    profile_name = input(f"🎤 Profile name [default]: ").strip() or "default"
    profile = Profile(profile_name)
    profile.setup()
    profile.save()

    settings = {
        "profile": profile_name,
        "last_update": dt.date.today().strftime("%y%m%d"),
        "use_wildmatch_ignore": False,
    }
    NSCSpecIO().write(settings)

    cprint("aok", "Initialisation complete!")

    if sys.version_info < (3, 7, 0):
        cprint(
            "war",
            "Support for Python 3.6 is deprecated, and will be "
            "removed in v1.2.",
        )


def setup(subparsers):
    subparsers.add_parser("init", description="Initialise nusex.")
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
import shutil
from pathlib import Path

from nusex import (
    CONFIG_DIR,
    LICENSE_DIR,
    PROFILE_DIR,
    TEMPLATE_DIR,
    Profile,
    __version__,
)
from nusex.errors import MigrationError
from nusex.helpers import cprint
from nusex.spec import NSCSpecIO, NSXSpecIO
from nusex.utils import Downloader


def _migrate():
    # Back-up
    shutil.copytree(CONFIG_DIR, f"{CONFIG_DIR}-old")

    # Create directories
    for dir in (LICENSE_DIR, PROFILE_DIR, TEMPLATE_DIR):
        os.makedirs(dir, exist_ok=True)

    # Download assets
    for dl in ("templates", "licenses"):
        Downloader(dl).download()

    # Create new profile
    profile = Profile.from_legacy()
    profile.save()

    # Convert templates
    for file in Path(CONFIG_DIR).glob("*.nsx"):
        with open(file) as f:
            data = json.load(f)

        for k, v in data["files"].items():
            data["files"][k] = v.encode()
        data.update({"installs": [], "as_extension_for": ""})
        NSXSpecIO().write(TEMPLATE_DIR / f"{file.stem}.nsx", data)

    # Remove old files
    for file in Path(CONFIG_DIR).glob("*.*"):
        os.remove(file)

    # Create config file
    settings = {
        "profile": "default",
        "last_update": __version__,
        "use_wildmatch_ignore": False,
    }
    NSCSpecIO().write(settings)


def _revert():
    shutil.rmtree(CONFIG_DIR)
    shutil.move(f"{CONFIG_DIR}-old", CONFIG_DIR)


def run(revert):
    if revert:
        if not (CONFIG_DIR.parent / f"{CONFIG_DIR}-old").exists():
            raise MigrationError("No old configurations to revert to")

        _revert()
        return cprint("aok", "Migration successfully reverted!")

    if not (CONFIG_DIR / "user.nsc").exists():
        raise MigrationError("Nothing to migrate from")

    try:
        _migrate()
        cprint("aok", "Migration successful!")
    except Exception as exc:
        _revert()
        cprint(
            "err",
            "Migration aborted. Your old configuration has been restored.",
        )
        raise exc


def setup(subparsers):
    s = subparsers.add_parser(
        "migrate", description="Migrate from a 0.x config to a 1.x one."
    )
    s.add_argument(
        "--revert",
        help="Revert back to an 0.x config, if possible.",
        action="store_true",
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex import PROFILE_DIR
from nusex.constants import TEMPLATE_DIR
from nusex.helpers import cprint


def run(profiles, templates):
    if not profiles and not templates:
        profiles = templates = True

    if profiles:
        cprint("inf", "Showing all profiles:")

        for p in sorted(PROFILE_DIR.glob("*.nsp")):
            print(p.stem)

        if templates:
            # Awkward lining up to make it prettier
            print()

    if templates:
        cprint("inf", "Showing all templates:")

        for t in sorted(TEMPLATE_DIR.glob("*.nsx")):
            print(t.stem)


def setup(subparsers):
    s = subparsers.add_parser(
        "list", description="Display a list of profiles and templates."
    )
    s.add_argument(
        "-p", "--profiles", help="only display profiles", action="store_true"
    )
    s.add_argument(
        "-t", "--templates", help="only display templates", action="store_true"
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex.errors import UnsupportedFile

SPEC_ID = b"\x99\x78"


class NSXSpecIO:
    __slots__ = ("defaults",)

    def __init__(self):
        self.defaults = {
            "files": {},
            "installs": [],
            "as_extension_for": "",
        }

    def _process_files(self, f, data):
        key = True
        name = b""

        while True:
            if key:
                chunk = f.read(1)

                if chunk == b"\x97":
                    key = False
                    continue

                if chunk == b"\x98":
                    return data

                name += chunk

            else:
                size = int(f.read(8).decode().strip(), base=16)
                chunk = f.read(size)
                data["files"].update({name.decode(): chunk})
                name = b""
                key = True

    def _process_installs(self, f, data):
        inst = b""
        while True:
            b = f.read(1)
            if b == b"\x98":
                return data

            if b == b"\x97":
                data["installs"].append(inst.decode())
                inst = b""
                continue

            inst += b

    def read(self, path):
        data = self.defaults.copy()
        with open(path, "rb") as f:
            # Validate format.
            if f.read(2) != SPEC_ID:
                raise UnsupportedFile("Not a valid NSX file")

            # Read headers.
            ef = f.read(1)
            if ef == b"\x01":
                data["as_extension_for"] = f.read(24).decode().strip()

            f.read(9)  # Skip reserved.

            # Process chunks.
            while f.peek(1):
                data = {
                    b"\x01": self._process_files,
                    b"\x02": self._process_installs,
                }[f.read(1)](f, data)

        return data

    def write(self, path, data):
        with open(path, "wb") as f:
            # Identify format.
            f.write(SPEC_ID)

            # Write headers.
            ef = data["as_extension_for"]
            if ef:
                f.write(b"\x01")
                f.write(ef.ljust(24).encode())
            else:
                f.write(b"\x00")

            f.write(b"\x00" * 9)  # Reserved space.

            # Files chunk starting byte.
            f.write(b"\x01")
            for k, v in data["files"].items():
                f.write(k.encode())
                f.write(b"\x97")
                size = len(v)
                if size > 0xFFFFFFFF:
                    raise UnsupportedFile(
                        "Files larger than 4 GB are not supported"
                    )
                f.write(hex(len(v))[2:].ljust(8).encode())
                f.write(v)
            f.write(b"\x98")

            # Installs chunk starting byte.
            f.write(b"\x02")
            for i in data["installs"]:
                f.write(i.encode())
                f.write(b"\x97")
            f.write(b"\x98")



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from nusex import PROFILE_DIR, TEMPLATE_DIR, Profile
from nusex.errors import DoesNotExist, ProfileError, TemplateError
from nusex.helpers import cprint, validate_name


def run(name, new_name):
    if name.startswith("nsx"):
        raise TemplateError("You cannot rename premade templates")

    if (PROFILE_DIR / f"{name}.nsp").exists():
        if Profile(name) == Profile.current():
            raise ProfileError(
                "You cannot rename the currently selected profile"
            )
        validate_name(new_name, "Profile")
        os.rename(PROFILE_DIR / f"{name}.nsp", PROFILE_DIR / f"{new_name}.nsp")
        cprint(
            "aok", f"Profile '{name}' successfully renamed to '{new_name}'!"
        )

    elif (TEMPLATE_DIR / f"{name}.nsx").exists():
        validate_name(new_name, "Template")
        os.rename(
            TEMPLATE_DIR / f"{name}.nsx", TEMPLATE_DIR / f"{new_name}.nsx"
        )
        cprint(
            "aok", f"Template '{name}' successfully renamed to '{new_name}'!"
        )

    else:
        raise DoesNotExist(f"No profile or template with name '{name}' found")


def setup(subparsers):
    s = subparsers.add_parser(
        "rename", description="Rename a profile or template."
    )
    s.add_argument(
        "name",
        help="the name of the profile or template to rename",
    )
    s.add_argument(
        "new_name",
        help="the new name for the profile or template",
    )
    return subparsers



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import datetime as dt
import sys
import traceback
from importlib import import_module
from pathlib import Path
from urllib import request
from urllib.error import HTTPError

from nusex import CONFIG_DIR, CONFIG_FILE, __description__, __version__
from nusex.errors import NusexError, NusexUserError
from nusex.helpers import cprint
from nusex.spec import NSCSpecIO
from nusex.utils import Downloader

LAST_UPDATE_URL = (
    "https://raw.githubusercontent.com/nusex/downloads/main/lastupdate.txt"
)
COMMAND_MAPPING = {
    p.stem: import_module(f".cli.commands.{p.stem}", package="nusex")
    for p in Path(__file__).parent.glob("commands/*.py")
    if p.stem != "__init__"
}

parser = argparse.ArgumentParser(description=__description__)
parser.add_argument(
    "-v",
    "--version",
    help="show nusex's version and exit",
    action="store_true",
)
subparsers = parser.add_subparsers(dest="subparser")
for module in COMMAND_MAPPING.values():
    subparsers = module.setup(subparsers)  # type: ignore


def _check_config(subcommand):
    if (CONFIG_DIR / "user.nsc").exists() and subcommand != "migrate":
        cprint(
            "err",
            "It looks like you still have an old nusex configuration. Use "
            "`nusex migrate` to fix this.",
        )
        sys.exit(2)


def _check_init(subcommand):
    if not CONFIG_FILE.exists() and subcommand not in ("init", "migrate"):
        cprint(
            "err",
            "That command cannot be run before nusex has been initialised.",
        )
        sys.exit(2)


def _check_for_updates():
    if not CONFIG_FILE.exists():
        return

    data = NSCSpecIO().read()
    last_checked = data["last_update"]
    auto_update = data["auto_update"]

    try:
        last_checked = dt.datetime.strptime(last_checked, "%y%m%d").date()
    except ValueError:
        # Probably some old config.
        last_checked = dt.date.min

    if last_checked == dt.date.today():
        return

    try:
        with request.urlopen(LAST_UPDATE_URL) as r:
            last_update = dt.datetime.strptime(
                r.readlines()[0].strip().decode(), "%y%m%d"
            ).date()
    except HTTPError:
        return

    if last_checked < last_update:
        if auto_update:
            cprint(
                "inf", "nusex is automatically downloading asset updates..."
            )
            Downloader("templates").download(display_progress=True)
            Downloader("licenses").download(display_progress=True)
        else:
            cprint(
                "inf",
                "nusex has asset updates. Use `nusex download` to get them.",
            )

    data["last_update"] = dt.date.today().strftime("%y%m%d")
    NSCSpecIO().write(data)


def main():
    args = parser.parse_args()

    # NOTE: Remove in v1.1.
    if parser.prog == "nsx":
        cprint(
            "war",
            (
                "The 'nsx' command is deprecated, and will be removed in v1.1 "
                "(use 'nusex' instead)."
            ),
        )

    if args.version:
        return print(__version__)

    if not args.subparser:
        return parser.parse_args(("-h",))

    # # Setup checks.
    _check_config(args.subparser)
    _check_init(args.subparser)
    _check_for_updates()

    # Command runs.
    try:
        COMMAND_MAPPING[args.subparser].run(
            **{
                k: v
                for k, v in args.__dict__.items()
                if k not in ("subparser", "version")
            }
        )
    except NusexUserError as exc:
        cprint("err", f"{exc}.")
        sys.exit(2)
    except NusexError as exc:
        cprint("err", f"{exc}.")
        sys.exit(1)
    except KeyboardInterrupt as exc:
        sys.exit(130)
    except Exception as exc:
        cprint(
            "err",
            f"Oh no! Something went wrong.\n\n{traceback.format_exc()}",
            end="",
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex.errors import UnsupportedFile

SPEC_ID = b"\x99\x70"


class NSPSpecIO:
    __slots__ = ("rmap", "wmap", "defaults")

    def __init__(self):
        self.rmap = {
            b"\x01": "author_name",
            b"\x02": "author_email",
            b"\x03": "git_profile_url",
            b"\x04": "starting_version",
            b"\x05": "default_description",
            b"\x06": "preferred_license",
        }
        self.wmap = {
            "author_name": b"\x01",
            "author_email": b"\x02",
            "git_profile_url": b"\x03",
            "starting_version": b"\x04",
            "default_description": b"\x05",
            "preferred_license": b"\x06",
        }
        self.defaults = {
            "author_name": "John Smith",
            "author_email": "thedoctor@email.com",
            "git_profile_url": "https://github.com/shakespearecode",
            "starting_version": "0.1.0",
            "default_description": "My project, made using nusex",
            "preferred_license": "mit",
        }

    def _scan(self, path):
        with open(path, "rb") as f:
            # Validate format.
            if f.read(2) != SPEC_ID:
                raise UnsupportedFile("Not a valid NSP file")

            while f.peek(1):
                key = self.rmap[f.read(1)]
                value = b""
                while True:
                    char = f.read(1)
                    if char == b"\x97":
                        break
                    value += char

                yield key, value.decode()

    def read(self, path):
        data = self.defaults.copy()
        data.update({k: v for k, v in self._scan(path)})
        return data

    def write(self, path, data):
        with open(path, "wb") as f:
            # Identify format.
            f.write(SPEC_ID)

            # Write data.
            for k, v in data.items():
                f.write(self.wmap[k])
                f.write(v.encode())
                f.write(b"\x97")



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

if sys.version_info < (3, 6, 0):
    print(
        "nusex only supports Python versions 3.6.0 or greater.",
        file=sys.stderr,
    )
    sys.exit(1)

import setuptools


def parse_requirements(path):
    with open(path) as f:
        deps = (d.strip() for d in f.readlines())
        return [d for d in deps if not d.startswith(("#", "-r"))]


with open("./nusex/__init__.py") as f:
    (
        productname,
        version,
        description,
        url,
        docs,
        author,
        email,
        license_,
        bug_tracker,
        ci,
    ) = [l.split('"')[1] for l in f.readlines()[28:38]]

with open("./README.md") as f:
    long_description = f.read()

setuptools.setup(
    name=productname,
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=url,
    author=author,
    author_email=email,
    license=license_,
    classifiers=[
        # "Development Status :: 1 - Planning",
        # "Development Status :: 2 - Pre-Alpha",
        # "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        # "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Utilities",
    ],
    project_urls={
        "Documentation": docs,
        "Source": url,
        "Bug Tracker": bug_tracker,
        "CI": ci,
    },
    python_requires=">=3.6.0,<3.12",
    packages=setuptools.find_packages(exclude=["tests*"]),
    entry_points={
        "console_scripts": [
            "nusex = nusex.cli.cli:main",
            "nsx = nusex.cli.cli:main",
        ]
    },
)



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime as dt
import json
import os
import sys
from pathlib import Path
from platform import python_implementation

from nusex import TEMP_DIR, TEMPLATE_DIR, Profile
from nusex.blueprints import PythonBlueprint
from nusex.constants import LICENSE_DIR
from nusex.errors import BuildError, IncompatibilityError, TemplateError
from nusex.helpers import run, validate_name
from nusex.spec import NSXSpecIO

ATTRS = (
    "PROJECTNAME",
    "PROJECTAUTHOR",
    "PROJECTAUTHOREMAIL",
    "PROJECTURL",
    "PROJECTVERSION",
    "PROJECTDESCRIPTION",
    "PROJECTLICENSE",
    "PROJECTYEAR",
    "LICENSEBODY",
    "PROJECTBASEEXC",
)


class Template:
    """A class in which to create, load, modify, and save templates.

    Args:
        name (str): The name of the template. If the template does not
            exist, a new one is created, otherwise an existing one is
            loaded.

    Keyword Args:
        installs (list[str]): A list of dependancies to be installed
            when the template is deployed.

    Attributes:
        path (pathlib.Path): The complete filepath to the template.
        data (dict[str, Any]): The data for the template.
    """

    __slots__ = ("path", "data", "_installs")

    def __init__(self, name, *, installs=[]):
        self.path = TEMPLATE_DIR / f"{name}.nsx"
        self._installs = installs

        if not os.path.isfile(self.path):
            return self.create_new(name)

        self.load()

    def __str__(self):
        return self.path.stem

    def __repr__(self):
        return (
            f"<Template name={self.name!r} files={len(self.data['files'])!r}>"
        )

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __getitem__(self, key):
        return self.data["files"][key]

    @property
    def name(self):
        """The name of the template.

        Returns:
            str
        """
        return self.path.stem

    @property
    def exists(self):
        """Whether the template exists on disk.

        Returns:
            bool
        """
        return self.path.is_file()

    def create_new(self, name):
        """Create a new template. This should never need to be
        called as templates are created automatically when necessary
        upon object instantiation.

        Args:
            name (str): The name of the template.

        Raises:
            TemplateError: The provided name is invalid.
            AlreadyExists: The template already exists on disk.
        """
        validate_name(name, self.__class__.__name__)
        self.data = {
            "files": {},
            "installs": [],
            "as_extension_for": "",
        }

    def load(self):
        """Load an existing template. This should never need to be
        called as templates are loaded automatically when necessary upon
        object instantiation.

        Raises:
            FileNotFoundError: The template does not exist on disk.
        """
        self.data = NSXSpecIO().read(self.path)

    def save(self):
        """Save this profile.

        Raises:
            TemplateError: The profile data has been improperly
                modified.
        """
        try:
            NSXSpecIO().write(self.path, self.data)
        except KeyError:
            raise TemplateError(
                "The template data has been improperly modified"
            ) from None

    def delete(self):
        """Delete this template.

        Raises:
            FileNotFoundError: The template does not exist on disk.
        """
        os.remove(self.path)

    def rename(self, new_name):
        """Rename this template.

        Args:
            new_name (str): The new name for the template.

        Raises:
            FileNotFoundError: The template does not exist on disk.
        """
        new_path = f"{self.path}".replace(self.path.stem, new_name)
        self.path = self.path.rename(new_path)

    @classmethod
    def from_cwd(
        cls, name, *, installs=[], ignore_exts=set(), ignore_dirs=set()
    ):
        """Create a template using files from the current working
        directory.

        Args:
            name (str): The name of the template.

        Keyword Args:
            installs (list[str]): A list of dependencies to install when
                this template is deployed. Defaults to an empty list.
            ignore_exts (set[str]): A set of file extensions to ignore.
                Defaults to an empty set.
            ignore_dirs (set[str]): A set of directories to ignore.
                Defaults to an empty set.

        Returns:
            Template: The newly created template.
        """
        c = cls(name, installs=installs)
        c.build(
            ignore_exts=ignore_exts,
            ignore_dirs=ignore_dirs,
        )
        return c

    @classmethod
    def from_dir(
        cls, name, path, *, installs=[], ignore_exts=set(), ignore_dirs=set()
    ):
        """Create a template using files from a specific directory.

        Args:
            name (str): The name of the template.
            path (str | os.PathLike): The path to the files to build the
                template with.

        Keyword Args:
            installs (list[str]): A list of dependencies to install when
                this template is deployed. Defaults to an empty list.
            ignore_exts (set[str]): A set of file extensions to ignore.
                Defaults to an empty set.
            ignore_dirs (set[str]): A set of directories to ignore.
                Defaults to an empty set.

        Returns:
            Template: The newly created template.
        """
        c = cls(name, installs=installs)
        c.build(
            root_dir=path,
            ignore_exts=ignore_exts,
            ignore_dirs=ignore_dirs,
        )
        return c

    @classmethod
    def from_repo(
        cls, name, url, *, installs=[], ignore_exts=set(), ignore_dirs=set()
    ):
        """Create a template using files from a GitHub repository.

        Args:
            name (str): The name of the template.
            url (str): The URL of the GitHub repository to clone.

        Keyword Args:
            installs (list[str]): A list of dependencies to install when
                this template is deployed. Defaults to an empty list.
            ignore_exts (set[str]): A set of file extensions to ignore.
                Defaults to an empty set.
            ignore_dirs (set[str]): A set of directories to ignore.
                Defaults to an empty set.

        Returns:
            Template: The newly created template.

        Raises:
            BuildError: Cloning the repository failed.
        """
        os.makedirs(TEMP_DIR, exist_ok=True)
        os.chdir(TEMP_DIR)

        output = run(f"git clone {url}")
        if output.returncode == 1:
            raise BuildError(
                "Cloning the repo failed. Is Git installed? Is the URL "
                "correct?"
            )

        os.chdir(TEMP_DIR / url.split("/")[-1].replace(".git", ""))
        return cls.from_cwd(
            name,
            installs=installs,
            ignore_exts=ignore_exts,
            ignore_dirs=ignore_dirs,
        )

    def get_file_listing(
        self, root_dir, *, ignore_exts=set(), ignore_dirs=set()
    ):
        """Get a list of files to include in this template.

        Args:
            root_dir (str): The root directory that nusex will search
                from.

        Keyword Args:
            ignore_exts (set[str]): A set of file extensions to ignore.
                Defaults to an empty set.
            ignore_dirs (set[str]): A set of directories to ignore.
                Defaults to an empty set.

        Returns:
            list[Path]: A list of filepaths.
        """

        def is_valid(path):
            return (
                path.is_file()
                and all(i not in path.parts for i in true_dir_ignores)
                and all(i[1:] not in f"{path}" for i in wild_dir_ignores)
                and all(i != path.suffix[1:] for i in ignore_exts)
            )

        wild_dir_ignores = set(
            filter(lambda x: x.startswith("*"), ignore_dirs)
        )
        true_dir_ignores = ignore_dirs - wild_dir_ignores
        files = filter(lambda p: is_valid(p), Path(root_dir).rglob("*"))
        return list(files)

    def build(self, project_name=None, files=[], root_dir=".", **kwargs):
        """Build this template. View the
        :doc:`template guide <../guide/templates>` to see what this
        command does in more detail.

        Keyword Args:
            project_name (str): The name of the project. If this is None
                the project name is set to the name of the parent
                folder. Defaults to None.
            files (list[str]): The list of files to include in this
                template. If no files are specified, the file listing
                is automatically retrieved.
            root_dir (str): The root directory that nusex will search
                from. Defaults to the current directory.
            **kwargs (Any): Arguments for the :code:`get_file_listing`
                method.
        """

        def resolve_key(path):
            path = "/".join(f"{path.resolve()}".split(os.sep)[nparts:])
            return path.replace(project_name, "PROJECTNAME")

        if not project_name:
            project_name = Path(root_dir).resolve().parts[-1]

        if not files:
            files = self.get_file_listing(
                root_dir,
                ignore_exts=kwargs.pop("ignore_exts", set()),
                ignore_dirs=kwargs.pop("ignore_dirs", set()),
            )

        nparts = len(Path(root_dir).resolve().parts)
        data = {
            "files": {resolve_key(f): f.read_bytes() for f in files},
            "installs": self._installs,
            "as_extension_for": "",
        }

        blueprint = PythonBlueprint(project_name, data)
        self.data = blueprint().data

    def deploy(self, path="."):
        """Deploy this template.

        Keyword Args:
            path (str): The path to deploy this template to. Defaults to
                the current directory.
        """

        def resolve_version(key):
            if key != "CALVER":
                return key

            return dt.date.today().strftime("%Y.%m.%d")

        def resolve_license_info(key):
            with open(LICENSE_DIR / f"{key}.txt", encoding="utf-8") as f:
                lines = f.read().split("\n")

            start = [i for i, line in enumerate(lines) if line == "---"][
                -1
            ] + 2

            return (
                lines[1][7:],
                "\n".join(lines[start:])
                .replace("[year]", f"{dt.date.today().year}")
                .replace("[fullname]", profile["author_name"]),
            )

        profile = Profile.current()
        project_name = Path(path).resolve().parts[-1]
        lic_name, lic_body = resolve_license_info(profile["preferred_license"])
        project_error = project_name.replace("_", " ").title().replace(" ", "")

        var_mapping = {
            b"PROJECTNAME": project_name,
            b"PROJECTVERSION": resolve_version(profile["starting_version"]),
            b"PROJECTDESCRIPTION": profile["default_description"],
            b"PROJECTURL": f"{profile['git_profile_url']}/{project_name}",
            b"PROJECTAUTHOREMAIL": profile["author_email"],
            b"PROJECTAUTHOR": profile["author_name"],
            b"PROJECTLICENSE": lic_name,
            b"LICENSEBODY": lic_body,
            b"PROJECTYEAR": f"{dt.date.today().year}",
            b"PROJECTBASEEXC": f"{project_error}Error",
        }

        for name, data in self.data["files"].items():
            name = name.replace(
                "PROJECTNAME",
                project_name.lower().replace(" ", "_").replace("-", "_"),
            )

            dirs = name.split("/")[:-1]
            if dirs:
                os.makedirs(f"{path}/" + "/".join(dirs), exist_ok=True)

            for k, v in var_mapping.items():
                data = data.replace(k, v.encode())

            with open(f"{path}/{name}", "wb") as f:
                f.write(data)

        meta = {
            "template": self.name,
            "files": list(self.data["files"].keys()),
        }
        with open(f"{path}/.nusexmeta", "w") as f:
            json.dump(meta, f)

    def install_dependencies(self):
        """Install this template's dependencies. Note that this does not
        work on PyPy Python implementations."""
        if python_implementation() == "PyPy":
            raise IncompatibilityError(
                "Dependency installation is not supported on PyPy "
                "implementations"
            )

        if not self.data["installs"]:
            return

        run(
            f"{sys.executable} -m pip install "
            + " ".join(self.data["installs"])
        )

    def check(self):
        """Check the template manifest, including line changes.

        Returns:
            dict[str, list[tuple[int, str]]]: The template manifest. The
            keys are always file names, and the values are tuples
            of the line numbers and line values that have been changed.
            This may not always be present.
        """
        manifest = {}

        for file, data in self.data["files"].items():
            manifest.update({file: []})

            try:
                for i, line in enumerate(data.decode().split("\n")):
                    if any(a in line for a in ATTRS):
                        manifest[file].append((i + 1, line))
            except UnicodeDecodeError:
                # If it errors here, no modifications could have been
                # made.
                ...

        return manifest



# Copyright (c) 2021, Ethan Henderson
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nusex import CONFIG_FILE
from nusex.errors import UnsupportedFile

SPEC_ID = b"\x99\x63"


class NSCSpecIO:
    __slots__ = ("defaults",)

    def __init__(self):
        self.defaults = {
            "profile": "default",
            "last_update": "000101",
            "use_wildmatch_ignore": False,
            "auto_update": False,
        }

    def read(self):
        data = self.defaults.copy()
        with open(CONFIG_FILE, "rb") as f:
            # Validate format.
            if f.read(2) != SPEC_ID:
                raise UnsupportedFile("Not a valid NSC file")

            # Load profile data.
            data["profile"] = f.read(24).decode().strip()
            date = f.read(6).decode()
            try:
                int(date)
                data["last_update"] = date
            except ValueError:
                # Some invalid or broken config.
                ...

            # Not guaranteed from here.
            attrs = ("use_wildmatch_ignore", "auto_update")
            for attr in attrs:
                try:
                    data[attr] = f.read(1) == b"\x01"
                except Exception as exc:
                    # Most likely no more options to read, so exit.
                    break

        return data

    def write(self, data):
        with open(CONFIG_FILE, "wb") as f:
            # Identify format.
            f.write(SPEC_ID)

            # Write data.
            f.write(data["profile"].ljust(24).encode())
            f.write(data["last_update"].encode())

            # Not guaranteed, so write a default value if not present.
            f.write((b"\x00", b"\x01")[data.get("use_wildmatch_ignore", 0)])
            f.write((b"\x00", b"\x01")[data.get("auto_update", 0)])
